"""
pyautogit.askpass module

This is an internal module meant to be used with pyautogit for git askpass git credential asking.

The two files are included to differentiate windows and linux, where on linux we use the python3 shebang, on windows the python shebang

(Hopefully with the death of python 2 we can only use one file)
"""